<?php

namespace Drp\NativeEventApp\Auth;

use DateInterval;
use DateTime;
use Drp\NativeEventApp\Auth\Facades\CodeGenerator;
use Laravel\Passport\AuthCode;
use League\OAuth2\Server\Entities\ClientEntityInterface;
use League\OAuth2\Server\Entities\UserEntityInterface;
use League\OAuth2\Server\Exception\OAuthServerException;
use League\OAuth2\Server\Exception\UniqueTokenIdentifierConstraintViolationException;
use League\OAuth2\Server\Grant\AbstractGrant;
use League\OAuth2\Server\Repositories\AuthCodeRepositoryInterface;
use League\OAuth2\Server\Repositories\RefreshTokenRepositoryInterface;
use League\OAuth2\Server\RequestEvent;
use League\OAuth2\Server\RequestTypes\AuthorizationRequest;
use League\OAuth2\Server\ResponseTypes\ResponseTypeInterface;
use LogicException;
use Psr\Http\Message\ServerRequestInterface;

/**
 * @SuppressWarnings(PHPMD.CouplingBetweenObjects)
 */
class AuthTokenGrant extends AbstractGrant
{
    /** @var DateInterval */
    private $authCodeTTL;

    public function __construct(
        AuthCodeRepositoryInterface $authCodeRepository,
        RefreshTokenRepositoryInterface $refreshTokenRepository,
        DateInterval $authCodeTTL,
        DateInterval $refreshTokenTTL
    ) {
        $this->setAuthCodeRepository($authCodeRepository);
        $this->setRefreshTokenRepository($refreshTokenRepository);
        $this->authCodeTTL = $authCodeTTL;
        $this->refreshTokenTTL = $refreshTokenTTL;
    }

    public function getIdentifier(): string
    {
        return 'authorization_token';
    }

    private function validateAuthorizationCode(AuthCode $authCode, ServerRequestInterface $request): void
    {
        if ($authCode->expires_at->isPast()) {
            throw OAuthServerException::invalidRequest('code', 'Authorization code has expired');
        }
        if ($this->authCodeRepository->isAuthCodeRevoked($authCode->id) === true) {
            throw OAuthServerException::invalidRequest('code', 'Authorization code has been revoked');
        }

        $authIdentifier = $this->getRequestParameter('auth_identifier', $request, null);
        if ((string) $authCode->user_id !== $authIdentifier) {
            throw OAuthServerException::invalidRequest('code', 'Authorization code was not issued to this user');
        }
    }

    /**
     * {@inheritdoc}
     */
    public function completeAuthorizationRequest(AuthorizationRequest $authorizationRequest)
    {
        if ($authorizationRequest->getUser() instanceof UserEntityInterface === false) {
            throw new LogicException('An instance of UserEntityInterface should be set on the AuthorizationRequest');
        }

        $authCode = $this->issueAuthCode(
            $this->authCodeTTL,
            $authorizationRequest->getClient(),
            $authorizationRequest->getUser()->getIdentifier(),
            $authorizationRequest->getRedirectUri(),
            $authorizationRequest->getScopes()
        );

        return new AuthTokenResponse($authCode);
    }

    protected function generateUniqueCode(): string
    {
        return CodeGenerator::authRequestCode();
    }

    /**
     * @param int|string $length
     */
    protected function generateUniqueIdentifier($length = 40): string
    {
        return CodeGenerator::authToken($length);
    }

    public function respondToAccessTokenRequest(
        ServerRequestInterface $request,
        ResponseTypeInterface $responseType,
        DateInterval $accessTokenTTL
    ): ResponseTypeInterface {
        // Validate request
        $client = $this->validateClient($request);
        $authCodeIdentifier = $this->getRequestParameter('code', $request, null);

        $authCode = AuthCode::find($authCodeIdentifier);

        if ($authCode === null) {
            throw OAuthServerException::invalidRequest('code');
        }

        try {
            $this->validateAuthorizationCode($authCode, $request);

            $scopes = $this->scopeRepository->finalizeScopes(
                $this->validateScopes(json_decode($authCode->scopes)),
                $this->getIdentifier(),
                $client,
                $authCode->user_id
            );
        } catch (LogicException $e) {
            throw OAuthServerException::invalidRequest('code', 'Cannot decrypt the authorization code', $e);
        }


        // Issue and persist new access token
        $accessToken = $this->issueAccessToken($accessTokenTTL, $client, $authCode->user_id, $scopes);
        $this->getEmitter()->emit(new RequestEvent(RequestEvent::ACCESS_TOKEN_ISSUED, $request));
        $responseType->setAccessToken($accessToken);

        // Issue and persist new refresh token if given
        $refreshToken = $this->issueRefreshToken($accessToken);

        if ($refreshToken !== null) {
            $this->getEmitter()->emit(new RequestEvent(RequestEvent::REFRESH_TOKEN_ISSUED, $request));
            $responseType->setRefreshToken($refreshToken);
        }

        // Revoke used auth code
        $this->authCodeRepository->revokeAuthCode($authCode->id);

        return $responseType;
    }

    /**
     * Issue an auth code.
     *
     * @param string                 $userIdentifier
     * @param string|null            $redirectUri
     * @param \League\OAuth2\Server\Entities\ScopeEntityInterface[] $scopes
     * @throws OAuthServerException
     * @throws UniqueTokenIdentifierConstraintViolationException
     * @return \League\OAuth2\Server\Entities\AuthCodeEntityInterface
     */
    protected function issueAuthCode( // phpcs:ignore
        DateInterval $authCodeTTL,
        ClientEntityInterface $client,
        $userIdentifier,
        $redirectUri,
        array $scopes = []
    ) {
        $authCode = $this->authCodeRepository->getNewAuthCode();
        $authCode->setExpiryDateTime((new DateTime())->add($authCodeTTL));
        $authCode->setClient($client);
        $authCode->setUserIdentifier($userIdentifier);

        foreach ($scopes as $scope) {
            $authCode->addScope($scope);
        }

        $maxGenerationAttempts = self::MAX_RANDOM_TOKEN_GENERATION_ATTEMPTS;
        while ($maxGenerationAttempts > 0) {
            $authCode->setIdentifier($this->generateUniqueCode());
            try {
                $this->authCodeRepository->persistNewAuthCode($authCode);

                return $authCode;
            } catch (UniqueTokenIdentifierConstraintViolationException $e) {
                if ($maxGenerationAttempts === 0) {
                    throw $e;
                }
            }
            $maxGenerationAttempts -= 1;
        }
    }
}
